import numpy as np
import math
from scipy import sparse

from xbanalysis.solvers.glpk_linear_problem import GlpkLinearProblem


class FbaProblem:

    def __init__(self, xba_model):
        """Initialize FbaProblem from a xba_model.

        raise an exceptions, if fba parameters are not configured in the xba_model

        FBA mass balanced constraints: S * v = 0
        - rows of stoichiometric matrix S:
          - all metabolites (incl. boundary/constant metabolites),
          - i.e. excluding enzymes
        - columns of stoichiometric matrix S:
          - all reactions not containing enzymes as reactants or products
          - i.e. exlcluding protein synthesis, protein degradation, enzyme transitions
          - reversible reactions are not getting split into forward and backward reaction (for now)

        FBA flux bounds:
        - lower and upper bounds for reactions extracted from xba_model
        - getter and setter methods

        FBA objective: c.T * v:
        - extracted from xba_model (fbcObjective), including direction
        - getter and setter methods

        :param xba_model: xba model with all required FBA parameters
        :type xba_model: XbaModel
        """
        self.xba_model = xba_model

        # identify reaction excluding protein synthesis and protein degradation
        self.rids = np.array([r.id for r in xba_model.reactions.values()
                              if r.sboterm.is_in_branch('SBO:0000205') is False and
                              r.sboterm.is_in_branch('SBO:0000179') is False])
        self.rid2idx = {mrid: idx for idx, mrid in enumerate(self.rids)}
        self.flux_bounds = np.array([[xba_model.reactions[rid].fbc_lb for rid in self.rids],
                                     [xba_model.reactions[rid].fbc_ub for rid in self.rids]])
        self.reversible = np.array([self.xba_model.reactions[rid].reversible for rid in self.rids])
        for idx in range(len(self.reversible)):
            if self.flux_bounds[0, idx] < 0 and np.logical_not(self.reversible[idx]):
                self.reversible[idx] = True
                print(f'reaction {idx} set to reversible')
        self.rids_rev = self.rids[self.reversible]
        self.rid2idx_rev = {rid: idx + len(self.rids) for idx, rid in enumerate(self.rids_rev)}

        used_sids = set()
        for rid in self.rids:
            for sid in xba_model.reactions[rid].reactants:
                used_sids.add(sid)
            for sid in xba_model.reactions[rid].products:
                used_sids.add(sid)
        self.sids = np.array(list(used_sids))
        self.sid2idx = {sid: idx for idx, sid in enumerate(self.sids)}

        # sparse matrices faster and more efficient to handle
        self.s_mat_coo = xba_model.get_stoic_matrix_coo(self.sids, self.rids)

        self.is_fba_model = self.check_fba_model(xba_model)
        if self.is_fba_model is False:
            print('Error: not a FBA model - missing parameters')
            return

        for oid, fbc_obj in xba_model.fbc_objectives.items():
            if fbc_obj.active is True:
                # self.obj_id = fbc_obj.id
                self.obj_dir = fbc_obj.direction
                self.objective = fbc_obj.coefficiants
                break

        # lp problem will only be created once required.
        self.lp = None

    def check_fba_model(self, xba_model):
        """Check if fba related data available in the XBA model

        :param xba_model: Xba model with data extracted from SBML file
        :type xba_model: XbaModel
        :return: indicator if model contains data required for FBA problem formulation
        :rtype: bool
        """
        is_fba_model = True

        r_check = {'flux lower bound': 'fbc_lb', 'flux upper bound': 'fbc_ub'}
        for check, param in r_check.items():
            missing = [rid for rid in self.rids if hasattr(xba_model.reactions[rid], param) is False]
            if len(missing) > 0:
                print(f'missing {check} in:', missing)
                is_fba_model = False

        if hasattr(xba_model, 'fbc_objectives') is False:
            print('missing flux objectives')
            is_fba_model = False
        else:
            active_objs = [oid for oid in xba_model.fbc_objectives if xba_model.fbc_objectives[oid].active]
            if len(active_objs) == 0:
                print('missing an active flux objective')
                is_fba_model = False

        return is_fba_model

    def update_reversible_rids(self):
        self.rids_rev = self.rids[self.reversible]
        self.rid2idx_rev = {rid: idx + len(self.rids) for idx, rid in enumerate(self.rids_rev)}

    def combine_fwd_rev(self, vector):
        """combine forward and backward flux information.

        :param vector: contains data on all reactions + backward information for reversible reactions
        :type vector: 1D ndarray of shape (len(mr_rids) + sum(reversible))
        :return: combined flux information
        :rtype: 1D ndarray of shape (len(mr_rids))
        """
        n_cols = len(self.rids)
        combined = vector[:n_cols]
        combined[self.reversible] -= vector[n_cols:]
        return combined

    def get_model_status(self):
        """ Return status on model

        :returns: status information in a dictionary
        :rtype: dict
        """
        status = {'dof': self.s_mat_coo.shape[1] - np.linalg.matrix_rank(self.s_mat_coo.todense()),
                  'n_rids': len(self.rids), 'n_sids': len(self.sids)}
        return status

    def create_core_fba_lp(self):
        """create a core Linear Programming Problem for FBA with split reactions

        Core Problem contains coefficient matrix, variable and row bounds
        Objective function yet needs to be added prior to any optimization

        Splitting reactions in forward/backward directions reduces high values during FBA
        configured bounds are >= 0
        Splitting required for pFBA (minimize absolute flux values)

        :return: LpProblem configured for FBA
        :rtype: LpProblem
        """
        fwd_rev_s_mat_coo = sparse.hstack([self.s_mat_coo,
                                           -self.s_mat_coo.tocsc()[:, self.reversible]])
        fwd_rev_col_bnds = np.hstack((np.fmax(self.flux_bounds[:], 0),
                                      np.fmax(np.flip(-self.flux_bounds[:, self.reversible], axis=0), 0)))
        row_bnds = np.zeros((fwd_rev_s_mat_coo.shape[0], 2))
        lp = GlpkLinearProblem(fwd_rev_s_mat_coo, row_bnds, fwd_rev_col_bnds)
        return lp

    def delete_fba_lp(self):
        """Delete the corresponding linear problem, when required"""
        if self.lp is not None:
            del self.lp
            self.lp = None

    def get_coefficients(self, objective):
        """get coefficients for objective or constraint function

        :param objective: objective/constraint coefficients
        :type objective: dict with key: reaction id, value: stoichiometric coefficient
        :return: coefficients aligned with the linear problem
        :rtype: 1D ndarray with float
        """
        coefs = np.zeros(len(self.rids) + len(self.rids_rev))
        for rid, coef in objective.items():
            coefs[self.rid2idx[rid]] = coef
            if rid in self.rids_rev:
                coefs[self.rid2idx_rev[rid]] = -coef
        return coefs

    def set_lp_objective(self, objective, direction='maximize'):
        """Set objective coefficients and direction in linear problem

        :param objective: objective coefficients
        :type objective: dict with key: reaction id, value: stoichiometric coefficient
        :param direction: direction of optimization
        :type direction: str (optional, default: 'maximize')
        """
        assert self.lp is not None
        self.lp.set_obj_coefs(self.get_coefficients(objective))
        self.lp.set_obj_dir(direction)

    def remove_lp_objective(self, objective):
        """remove objective coefficients

        :param objective: objective coefficients
        :type objective: dict with key: reaction id, value: stoichiometric coefficient
        """
        self.lp.remove_obj_coefs(self.get_coefficients(objective))

    def update_lp_flux_bounds(self, flux_bounds):
        """update flux bounds for selected reactions in the linear problem.

        lp problem might need to be instantiated
        overwrite bounds of protected functions will be cleaned from already removed reactions

        :param flux_bounds: lower and upper flux bounds for selected reactions
        :type flux_bounds: dict: key: reaction id, value: list-like of two float values
        """
        if self.lp is None:
            self.lp = self.create_core_fba_lp()
        for rid, bound_upd in flux_bounds.items():
            if rid in self.rids:
                temp_bounds = self.flux_bounds[:, self.rid2idx[rid]].copy()
                for i in [0, 1]:
                    if math.isfinite(bound_upd[i]):
                        temp_bounds[i] = bound_upd[i]
                self.lp.set_col_bnd(self.rid2idx[rid], np.fmax(temp_bounds, 0))
                if rid in self.rids_rev:
                    self.lp.set_col_bnd(self.rid2idx_rev[rid], np.fmax(np.flip(-temp_bounds), 0))

    def reset_lp_flux_bounds(self, rids):
        """restore original flux bounds for selected reactions in linear problem.

        overwrite bounds of protected functions will be cleaned from already removed reactions

        :param rids: reaction ids for which to restore the model orignal flux bounds
        :type rids: list-like of str
        """
        assert self.lp is not None

        for rid in rids:
            if rid in self.rids:
                assert rid in self.rids
                bounds = self.flux_bounds[:, self.rid2idx[rid]]
                self.lp.set_col_bnd(self.rid2idx[rid], np.fmax(bounds, 0))
                if rid in self.rids_rev:
                    self.lp.set_col_bnd(self.rid2idx_rev[rid], np.fmax(np.flip(-bounds), 0))

    def fba_optimize(self, short_result=False):
        """FBA optimization of current problem

        :param short_result: short result (only objective value and status) or long result
        :type short_result: bool (optional, default: False)
        :return: results information from the optimization
        :rtype: dict
        """
        if self.lp is None:
            self.lp = self.create_core_fba_lp()
            # print('Core LP Problem created')
        assert self.lp is not None

        self.set_lp_objective(self.objective, self.obj_dir)
        res = self.lp.solve(short_result)
        # print(res['success'], res['fun'])
        self.remove_lp_objective(self.objective)

        if short_result is False:
            res['x'] = self.combine_fwd_rev(res['x'])
            res['reduced_costs'] = self.combine_fwd_rev(res['reduced_costs'])
        return res

    def pfba_optimize(self, frac_of_opt=1.0):
        """pFBA parsimoneous flux balance analysis.

        0 ≤ frac_of_opt ≤ 1.0  (for minimization problems. frac_of_opt > 0)
        for maximizing objective, this sets lower bound to (frac_of_opt * optimum)
        for minimization objective, this sets upper bound to (optimum / frac_of_opt)

        :param frac_of_opt: objective value within
        :type frac_of_opt: float [0 ... 1] , optional (default: 1.0)`
        :return:
        :rtype: dict
        """
        # get optimum value for model objective function
        res = self.fba_optimize(short_result=True)
        if res['success'] is False:
            return res

        # implement model main objective as a constraint
        if self.obj_dir == 'maximize':
            row_bnds = np.array([res['fun'] * frac_of_opt, np.nan])
        else:
            row_bnds = np.array([np.nan, res['fun'] / frac_of_opt])
        row_idx0 = self.lp.add_constraint(self.get_coefficients(self.objective), row_bnds)

        # create pFBA objective (all ones)
        pfba_obj_coefs = np.ones(len(self.rids) + len(self.rids_rev))
        self.lp.set_obj_coefs(pfba_obj_coefs)
        self.lp.set_obj_dir('minimize')
        res = self.lp.solve(short_result=False)
        # remove pFBA objective
        self.lp.remove_obj_coefs(pfba_obj_coefs)
        # remove main model objective from constraints matrix
        self.lp.del_constraint(row_idx0)
        res['x'] = self.combine_fwd_rev(res['x'])
        res['reduced_costs'] = self.combine_fwd_rev(res['reduced_costs'])
        return res

    def cyclefree_fba_optimize(self):
        """cycle free FBA (remove loops from FBA solution)

        Reference:
        Desouki et al., 2015, CycleFreeFlux: efficient removal of thermodynamically
        infeasible loops from flux distributions

        Method:
        - run FBA and determine a flux distribution
        - implement model objective function as equality constraint with FBA optimum
        - configure a new objective function ('minimize' absolut fluxes)
            - exchange reactions are not considered and fluxes are fixed at FBA level
            - reactions with positive FBA fluxes get a coefficient of 1.0 and fluxes bound between zero and FBA level
            - reactions with negative FBA fluxes get a coefficient of -1.0 and fluxes bound between FBA level and zero
                - i.e. internal reactions are only allowed to shrink, but not change direction
            - run FBA on new objective and retrieve new flux values for reactions.
                - use new flux values with model objective value as optimum
        :return: result of cycleFree FBA solution
        :rtype: dict
        """
        # get optimum value for model objective function
        res = self.fba_optimize(short_result=False)
        if res['success'] is False:
            return res
        optimum = res['fun']

        cff_objective = {}
        cff_flux_bounds = {}
        for rid, idx in self.rid2idx.items():
            flux = res['x'][idx]
            if len(self.xba_model.reactions[rid].products) == 0:
                cff_flux_bounds[rid] = [flux, flux]
            elif flux >= 0:
                cff_flux_bounds[rid] = [0.0, flux]
                cff_objective[rid] = 1.0
            else:
                cff_flux_bounds[rid] = [flux, 0.0]
                cff_objective[rid] = -1.0

        # implement model main objective as an equality constraint with optimal value
        row_bnds = np.array([optimum, optimum])
        row_idx0 = self.lp.add_constraint(self.get_coefficients(self.objective), row_bnds)

        # configure cyclefree FBA objective
        self.set_lp_objective(cff_objective, 'minimize')
        self.update_lp_flux_bounds(cff_flux_bounds)
        res = self.lp.solve(short_result=False)
        self.reset_lp_flux_bounds(set(cff_flux_bounds))
        self.remove_lp_objective(cff_objective)
        self.lp.del_constraint(row_idx0)

        res['fun'] = optimum
        res['x'] = self.combine_fwd_rev(res['x'])
        res['reduced_costs'] = self.combine_fwd_rev(res['reduced_costs'])
        return res

    def fva_optimize(self, rids=None, frac_of_opt=1.0):
        """Flux variability analyzis on reactions within fraction of optimum

        0 ≤ frac_of_opt ≤ 1.0  (for minimization problems. frac_of_opt > 0)
        for maximizing objective, this sets lower bound to (frac_of_opt * optimum)
        for minimization objective, this sets upper bound to (optimum / frac_of_opt)

        :param rids: reaction ids to determine flux ranges
        :type rids: list-like of str, optional (default: None, all reactions)
        :param frac_of_opt: objective value within
        :type frac_of_opt: float [0 ... 1] , optional (default: 1.0)
        :return: rids, min and max flux values
        :rtype: dict
        """
        # get optimum value for model objective function
        res = self.fba_optimize(short_result=True)
        if res['success'] is False:
            return res

        # implement model main objective as a constraint
        if self.obj_dir == 'maximize':
            row_bnds = np.array([res['fun'] * frac_of_opt, np.nan])
        else:
            row_bnds = np.array([np.nan, res['fun'] / frac_of_opt])
        row_idx0 = self.lp.add_constraint(self.get_coefficients(self.objective), row_bnds)

        if rids is None:
            rids = self.rids
        flux_ranges = np.zeros((4, len(rids)))
        for idx, rid in enumerate(rids):
            # determine minimal flux for reaction rid
            self.set_lp_objective({rid: 1}, 'minimize')
            res = self.lp.solve(short_result=True)
            if res['success']:
                flux_ranges[0, idx] = res['fun']
                flux_ranges[2, idx] = self.lp.get_row_value(row_idx0)
            # determine maximal flux for reaction rid
            self.lp.set_obj_dir('maximize')
            res = self.lp.solve(short_result=True)
            if res['success']:
                flux_ranges[1, idx] = res['fun']
                flux_ranges[3, idx] = self.lp.get_row_value(row_idx0)
            # remove objective for reaction rid
            self.remove_lp_objective({rid: 1})
        # remove main model objective from constraints matrix
        self.lp.del_constraint(row_idx0)

        res = {'rids': rids, 'min': flux_ranges[0], 'max': flux_ranges[1],
               'obj_min': flux_ranges[2], 'obj_max': flux_ranges[3]}
        return res

    def copy(self):
        """Copy the FbaModel instance (while removing linear problem).

        A shallow copy, to support multiprocessing of FVA,
        which do not impact FBA model attributes, except for linear problem
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.__dict__['lp'] = None
        return result

    def __del__(self):
        # print('deletion of fba_problem')
        del self.lp
